#!/usr/bin/env node

let LB = require('./linebuffer');
let net = require('net');

function encode(obj) {
    return JSON.stringify(obj) + "\n";
}

function decode(str) {
    return JSON.parse(str);
}

class DB {
    constructor(host,port) {
        if (host) {
            this.open(host, port || 3000);
        }
    }

    init(conn) {
        this.calls = {};
        this.next = 0;
        this.conn = conn;

        new LB(conn, line => {
            let json = decode(line);
            let call = json.call;
            let { fn, err } = this.calls[call];
            if (json.error) {
                err(json.error);
            } else {
                fn(json);
            }
            if (!json.continue) {
                delete this.calls[call];
            }
        });
    }

    close() {
        if (this.conn) {
            try { this.conn.end(); } catch (e) { console.log({e}) }
            this.conn = undefined;
        }
    }

    send(obj, fn, err) {
        if (this.conn) {
            let call = this.next++;
            this.calls[call] = {fn, err};
            this.conn.write(encode({call, ...obj}));
        } else {
            if (this.name && this.host) {
                if (this.defer) {
                    return this.defer.push({obj, fn, err});
                } else {
                    let client = this;
                    let defer = this.defer = [{obj, fn, err}];
                    return this.open().catch(error => {
                        // fail any in-flight queries
                        for (let rec of defer) {
                            try { rec.err({error}) } catch (e) { }
                        }
                        client.defer = undefined;
                    });
                }
            }
            err(new Error("not connected"));
        }
    }

    open(host = this.host, port = this.port) {
        return new Promise((resolve, reject) => {
            let conn = net.connect({host, port});
            conn
                .on('error', error => {
                    this.defer = undefined;
                    // console.log({error, calls:this.calls});
                    reject(error);
                })
                .on('close', close => {
                    // console.log({close});
                    this.close();
                })
                .on('connect', connect => {
                    // console.log({connect});
                    this.init(conn);
                    // recover last database and auth on re-open
                    if (this.name && this.host === host) {
                        let promise = this.user ?
                            this.auth(this.user, this.pass).then(() => {
                                return this.use(this.name);
                            }) :
                            this.use(this.name);
                        promise.then(() => {
                            let defer = this.defer;
                            this.defer = undefined;
                            if (defer) {
                                for (let rec of defer) {
                                    this.send(rec.obj, rec.fn, rec.err);
                                }
                            }
                        }).then(() => {
                            resolve();
                        }).catch(error => {
                            this.defer = undefined;
                            reject(error);
                        });
                    } else {
                        this.host = host;
                        this.port = port;
                        resolve();
                    }
                });
        });
    }

    stat() {
        return new Promise((resolve, reject) => {
            this.send({stat: true}, resolve, reject)
        });
    }

    auth(user, pass) {
        return new Promise((resolve, reject) => {
            this.send({auth: user, pass}, reply => {
                this.user = user;
                this.pass = pass;
                resolve(reply);
            }, reject)
        });
    }

    use(name, opt) {
        return new Promise((resolve, reject) => {
            this.name = name;
            this.send({use: name, opt}, resolve, reject);
        });
    }

    get(key, opt = {}) {
        return new Promise((resolve, reject) => {
            this.send({get: key}, opt.raw ? resolve : reply => {
                if (reply.error) {
                    reject(reply.error);
                } else {
                    resolve(reply.value);
                }
            }, reject);
        });
    }

    put(key, value, opt = {}) {
        return new Promise((resolve, reject) => {
            this.send({put: key, value}, opt.raw ? resolve : reply => {
                if (reply.error) {
                    reject(reply.error);
                } else {
                    resolve(key, value);
                }
            }, reject);
        });
    }

    del(key, opt = {}) {
        return new Promise((resolve, reject) => {
            this.send({del: key}, opt.raw ? resolve : reply => {
                if (reply.error) {
                    reject(reply.error);
                } else {
                    resolve(key);
                }
            }, reject);
        });
    }

    list(opt, fn, op2 = {}) {
        return new Promise((resolve, reject) => {
            let call = this.next++;
            let recs = fn ? undefined : [];
            this.send({list: true, opt}, reply => {
                let { list, error } = reply;
                if (list || error) {
                    if (fn) {
                        if (op2.raw) {
                            fn(list);
                        } else if (list.key) {
                            fn(list.key, list.value);
                        } else {
                            fn(list);
                        }
                    } else {
                        recs.push(op2.raw ? reply : reply.list);
                    }
                }
                if (!reply.continue) {
                    resolve(recs);
                }
            }, reject);
        });
    }
}

if (module.parent) {
    module.exports = DB;
} else {
    let args = require('minimist')(process.argv.slice(2));
    let host = args.host || "localhost";
    let port = parseInt(args.port || 3000);
    let user = args.user || undefined;
    let pass = args.pass || undefined;
    let base = args.base || undefined;
    let db = new DB();

    let promise;

    function onerror(error) {
        console.log({db_error: error});
    }

    if (host) {
        promise = db.open(host, port).catch(onerror);
        if (user) {
            promise = promise.then(() => {
                return db.auth(user,pass);
            }).catch(onerror);
        }
        if (base) {
            promise = promise.then(() => {
                return db.use(base);
            }).catch(onerror);
        }
    }

    function prompt() {
        process.stdout.write(`:: `);
    }

    if (promise) {
        promise.then(prompt);
    } else {
        prompt();
    }

    new LB(process.stdin, line => {
        line = line.toString().trim().split(' ').map(v => v.trim());
        let cmd = line.shift();
        switch (cmd) {
            case 'close':
                db.close();
                prompt();
                break;
            case 'open':
                db.open(line[0] || db.host || host, parseInt(line[1] || db.port || port)).then(reply => {
                    prompt();
                }).catch(error => {
                    console.log({open_error: error});
                    prompt();
                });
                break;
            case 'stat':
                db.stat().then(reply => {
                    console.log(reply.stat);
                    prompt();
                }).catch(error => {
                    console.log({stat_error: error});
                    prompt();
                });
                break;
            case 'auth':
                db.auth(line[0], line[1]).then(reply => {
                    prompt();
                }).catch (error => {
                    console.log({auth_error: error});
                    prompt();
                });
                break;
            case 'use':
                db.use(line[0]).then(reply => {
                    prompt();
                }).catch (error => {
                    console.log({use_error: error});
                    prompt();
                });
                break;
            case 'get':
                db.get(line[0]).then(reply => {
                    console.log(reply);
                    prompt();
                }).catch(error => {
                    console.log({get_error: error});
                    prompt();
                });
                break;
            case 'put':
                let [ key, val ] = line;
                if (val.charAt(0) === '{' || val.charAt(0) === '[') {
                    val = eval(`(${val})`);
                }
                db.put(key, val).then(reply => {
                    prompt();
                }).catch(error => {
                    console.log({put_error: error});
                    prompt();
                });
                break;
            case 'del':
                db.del(line[0]).then(reply => {
                    prompt();
                }).catch(error => {
                    console.log({del_error: error});
                    prompt();
                });
                break;
            case 'list':
            case 'keys':
                let isKeys = cmd === 'keys';
                db.list({
                    gte: line[0],
                    lt: line[1],
                    values: isKeys ? false : true
                }, (key, value) => {
                    console.log(isKeys ? key : {key, value});
                }).then(() => {
                    prompt();
                }).catch(error => {
                    console.log({list_error: error});
                    prompt();
                });
                break;
            case '':
                prompt();
                break;
            default:
                console.log({invalid_command: cmd});
                prompt();
                break;
        }
    });
}
